"""Top-level package for Python ARTMAP."""

__author__ = """David Vinicius"""
__email__ = 'de.175@hotmail.com'
__version__ = '0.1.0'

from .python_artmap import *
from .art import ART
from .art_fuzzy import ARTFUZZY
from .artmap_fuzzy import ARTMAPFUZZY



import numpy as np
from python_artmap import ART, ARTFUZZY

class ARTMAPFUZZY(ART):    
    rho        = 0
    WAB        = []
    championsA = []


    def __init__(self, INPUT, OUTPUT, rhoARTa=0.5, rhoARTb=0.5, rhoInterART=0.5, alphaARTa=0.001, betaARTa=1, alphaARTb=0.001, betaARTb=1, maxValueArta=1, maxValueArtb=1):
        self.ArtA = ARTFUZZY(self.layerF0(INPUT, maxValueArta), rho=rhoARTa, alpha=alphaARTa, beta=betaARTa)
        self.ArtB = ARTFUZZY(self.layerF0(OUTPUT, maxValueArtb), rho=rhoARTb, alpha=alphaARTb, beta=betaARTb)
        
        self.rho  = rhoInterART
        self.WAB  = np.ones([1, OUTPUT.shape[0]])

    def train(self):
        interator = 0
        for inputB in self.ArtB.I:            
            categories      = self.ArtB.categories(inputB, self.ArtB.W)
            champion        = categories.max()
            championIndexB  = categories.argmax()

            while champion != 0:                            
                if self.ArtB.hadRessonance(inputB, self.ArtB.W[championIndexB]):                    
                    self.ArtB.W[championIndexB]    = self.ArtB.learn(inputB, self.ArtB.W[championIndexB])                                    
                    self.ArtB.activate(championIndexB)

                    break
                else:                    
                    categories[championIndexB] = 0
                    champion                   = categories.max()
                    championIndexB             = categories.argmax()
            else:                      
                self.ArtB.setW(np.insert(self.ArtB.W, len(self.ArtB.W), inputB, 0))                            
                championIndexB += 1
                self.ArtB.activate(championIndexB)
            
            for inputA in self.ArtA.I[interator:]:
                categories      = self.ArtA.categories(inputA, self.ArtA.W)
                championA       = categories.max()
                championIndexA  = categories.argmax()                

                while championA != 0:                                    
                    if self.ArtA.hadRessonance(inputA, self.ArtA.W[championIndexA]):
                                                
                        if self.hadRessonance(self.ArtB.Y[championIndexB], self.WAB[championIndexA], self.rho):
                            self.ArtA.W[championIndexA]  = self.ArtA.learn(inputA, self.ArtA.W[championIndexA])
                            self.WAB[championIndexA]     = self.activate(self.WAB[championIndexA], championIndexB)
                            break

                        else:
                            categories[championIndexA] = 0                
                            championA                  = categories.max()
                            championIndexA             = categories.argmax()

                            x                          = self.AND(inputA, self.ArtA.W[championIndexA])
                            newRho                     = (sum(x) / sum(inputA))                            

                            self.ArtA._rho            = newRho                        
                    else:
                        categories[championIndexA] = 0                
                        championA                  = categories.max()
                        championIndexA             = categories.argmax()
                else:
                    self.ArtA.setW(
                        np.insert(self.ArtA.W, len(self.ArtA.W), inputA, 0)
                    )                    
                    self.ArtA.activate(championIndexA+1)
                    self.WAB = np.insert(self.WAB, len(self.WAB), self.activate(self.WAB[championIndexA], championIndexB), 0)                    
                                                                    
                
                interator += 1
                break
        
    def activate(self, W, i):
        temp    = np.zeros(len(W))
        temp[i] = 1
        return list(temp)
    
    def test(self, INPUT, rho):  
        INPUT           = np.divide(INPUT, 1)                         
        INPUT           = np.concatenate((INPUT, (1-INPUT)), axis=0)
        categories      = self.ArtA.categories(INPUT, self.ArtA.W)
        championA       = categories.max()
        championIndexA  = categories.argmax()

        while championA != 0:
            if self.hadRessonance(INPUT, self.ArtA.I[championIndexA], rho):
                t    = list(self.WAB[championIndexA])
                artB = list(self.ArtB.W[t.index(1)])                
                s    = [str(i) for i in artB]
                return {
                    "index": t.index(1),
                    #"value": self.WAB[championIndexA],
                    "ArtB": artB,
                    "id": "".join(s).replace(".", "")
                }
            else:
                categories[championIndexA] = 0                
                championA                  = categories.max()
                championIndexA             = categories.argmax()
        
        return -1

            

    



import numpy as np

class ART():
    Y          = []
    Js         = []
    _W         = []    
    
    def __init__(self, I, alpha = 0.001, rho = 0.5, beta = 1):
        self._alpha = alpha
        self._rho   = rho
        self._beta  = beta
        self.setI(I)
        self.setW(np.ones((1, len(self.I[0]))))        

    @staticmethod
    def layerF0(I, valueMax = 0):
        IC = np.asarray(I)
        IC = ART.normalize(IC, valueMax)
        IC = ART.complement(IC)    
        return IC
    
    @staticmethod
    def normalize(arr, valueMax = 0):
        if valueMax == 0:
            valueMax = arr.max()
        
        if valueMax == 0:
            valueMax = 1

        I  = np.divide(arr, valueMax)
        return I
    
    @staticmethod
    def complement(I):        
        I = np.concatenate((I, (1-I)), axis=1)
        return I        
        

    def AND(self, arr1, arr2):        
        try:        
            return np.minimum(arr1, arr2)
        except Exception as e:            
            print("AND", arr1, arr2)
            quit()
    
    @property
    def I(self):
        return self._I
    
    
    def setI(self, I):
        if isinstance(I, np.ndarray):
            self._I = I
        else:
            self._I = np.array(I)
    
    @property
    def W(self):
        return self._W
    
    
    def setW(self, W):
        if isinstance(W, np.ndarray):
            self._W = W
        else:
            self._W = np.array(W)

    def hadRessonance(self, IC, W, rho):
        try:
            IC  = np.asarray(IC)        
            x   = np.asarray(self.AND(IC, W))
            y   = x.sum(axis=0) / IC.sum(axis=0)
            return (y >= rho)
        except Exception as e:
            print(e)
            print("HAD", IC, W)
            quit()
    
    def vigilanceValue(self, IC, W):
        x   = self.AND(IC, W)
        return (sum(x) / sum(IC))

    def categories(self, I, W):
        x = np.minimum(I, W)
        categories = x.sum(axis=1) / (self._alpha + W.sum(axis=1))
        return categories



import numpy as np
from python_artmap import ART


class ARTFUZZY(ART):

    championIndex = 0
    championValue = 0
    categoriesArray = []

    def __init__(self, I, alpha=0.001, rho=0.5, beta=1):
        super().__init__(I, alpha, rho, beta)

    def learn(self, IC, W):
        temp1 = self._beta * self.AND(IC, W)
        temp2 = (1 - self._beta) * IC
        return temp1 + temp2

    def activate(self, i):
        temp    = np.zeros(len(self.I))
        temp[i] = 1
        self.Y.append(list(temp))

    def hadRessonance(self, IC, W):
        x = self.AND(IC, W)
        return ((sum(x) / sum(IC)) >= self._rho)    

    def train(self):        
        for i in range(0, len(self.I)):
            self.match(i)

    def match(self, indexOfInput):
        categories    = self.categories(self.I[indexOfInput], self.W)        
        champion      = categories.max()
        championIndex = categories.argmax()

        while champion != 0:            
            if self.hadRessonance(self.I[indexOfInput], self.W[championIndex]):                
                self.W[championIndex] = self.learn(self.I[indexOfInput], self.W[championIndex])

                self.activate(championIndex)
                self.Js.append([indexOfInput, championIndex])

                self.championIndex = championIndex
                self.championValue = champion

                break
            else:
                categories[championIndex] = 0
                champion = categories.max()
                championIndex = categories.argmax()
        else:
            self.setW(
                np.insert(self.W, len(self.W), self.I[indexOfInput], 0)
            )
            self.Js.append([indexOfInput, championIndex+1])
            self.activate(championIndex+1)
            self.championIndex = championIndex+1
            self.championValue = champion

    '''def searchForChampions(self, indexOfInput):
        self.categoriesArray = self.categories(self._alpha)
        champion = max(self.categoriesArray)
        championIndex = self.categoriesArray.index(champion)

        while champion != 0:
            if self.hadRessonance(self.I[indexOfInput], self.W[championIndex]):

                self.championIndex = championIndex
                self.championValue = champion

                break
            else:
                self.categoriesArray[championIndex] = 0
                champion = self.valueOfChampion(self.categoriesArray)
                championIndex = self.indexOfChampion(self.categoriesArray)'''
